with ada.strings.unbounded;
use ada.strings.unbounded;
with Mengen; use Mengen;

package Grammatiken is

   -- Zur Modellierung von Grammatiken:
   -- ---------------------------------  
   --
   -- Der Typ Grammatik stellt eine besondere Form von Typ-2-Grammatiken dar: 
   -- Neben den gewhnlichen Alternativen der Form "A -> X1 X2 ... Xn" gibt
   -- es hier "Listen-Alternativen" der Form "A -> X1 ... Xn { Y1 .. Ym }",
   -- bei denen der Teil "Y1 .. Ym" innerhalb der geschweiften Klammern null- 
   -- oder mehrfach an den Alternativenanfang "X1 ... Xn" angehngt werden
   -- kann. Es gilt jeweils n >= 0 und m > 0.
   -- 
   -- Die Grammatiksymbole werden durch ganze Zahlen codiert:
   -- 
   --    + das Startsymbol S durch die Zahl 1
   --    + die Nichtterminale NT durch das Intervall [1 .. anzNT]
   --    + die Terminale T durch das Intervall [anzNT+1 .. anzV]
   --    + das Endesymbol # durch anzV+1
   --  
   ----------------------------------------------------------------
   
   -- Eine Alternative enthlt die Zeichenfolge in einem IntegerArray
   -- rhs passender Laenge, daneben die Codierung lhs der linken Seite
   -- der Regel sowie eine Zahl-Komponente wdh, die angibt, ob es sich 
   -- um eine gewhnliche Alternative (wdh = 0) oder um eine Listen-
   -- Alternative (wdh > 0) handelt und bei wdh > 0 den Index des ersten 
   -- Symbols (Y1) der Wiederholungsteils bezeichnet.
   
   -- Die Komponenten LL dient zur Speicherung der LL-Vorschau zu jeder
   -- Alternative (wird spter mit bestimmeInfos berechnet).
   -- Es handelt sich um eine Menge von terminalen Zeichen,
   -- die zustzlich auch das Endezeichen enthalten kann.
   
   -- Die Komponenten LL2 (wird ebenfalls mit bestimmeInfos berechnet)
   -- speichert die LL-Vorschau fr den Wiederholungsteil, d.h. fr
   -- Alternativen mit wdh > 0.

   type IntegerArray is array (Integer range <>) of Integer;
 
   type Alternativen_Rec (Laenge : Natural) is record
      lhs : Natural;  
      rhs : IntegerArray (1 .. Laenge);
      wdh : Natural := 0;
      LL  : Menge; 
      LL2 : Menge;
   end record;   
   
   type Alternative is access Alternativen_Rec;
   
   ----------------------------------------------------------------
   
   -- Eine Regel enthlt die zu einem Nichtterminal gehrigen Alternativen
   -- in einem AlternativenArray mit Anzahl vielen Eintrgen, wobei die
   -- Anzahl als Diskriminante mitgegeben wird.
      
   type AlternativenArray is array (Integer range <>) of Alternative;
 
   type Regel_Rec (Anzahl : Natural) is record
      alt : AlternativenArray (1 .. Anzahl) :=
               (others => null);
   end record;
   
   type Regel is access Regel_Rec;
   
   ----------------------------------------------------------------
   
   -- Eine Grammatik wird durch einen Verbund dargestellt, wobei
   -- anzNT und anzV als Diskriminanten auftreten. Die Klartexte der 
   -- Grammatiksymbole stehen innerhalb der Grammatik in einer 
   -- StringZeigerArray-Komponente text, das Ende-Symbol in einer
   -- gleichnamigen Komponente, ebenso die Regeln. Die Komponente
   -- name enthlt den Namen der Grammatik.
   
   -- Die Komponente eps dient zur Speicherung der epsilon-Symbole
   -- der Grammatik, die in bestimmeInfos berechnet werden.
   -- Entsprechend kann die Komponente firstM fr jedes NT die
   -- zugehrige first-Menge aufnehmen und die Komponente follow
   -- fr jedes NT die zugehrige follow-Menge. Die first- und
   -- follow-Mengen sind Mengen von terminalen Zeichen, die 
   -- zustzlich auch das Endezeichen enthalten knnen. 
   -- Auch diese Mengen werden in bestimmeInfos berechnet.
   
   type StringArray is array (Integer range <>) of Unbounded_String;
   
   type RegelArray is array (Integer range <>) of Regel;
   
   type MengenArray is array (Integer range <>) of Menge;
   
   type Grammatik_Rec (anzNT,anzV : Positive) is record 
      name   : Unbounded_String;
      text   : StringArray (1 .. anzV); 
      ende   : Positive := anzV + 1;
      regeln : RegelArray (1 .. anzNT) := (others => null);
      eps    : Menge := LeerMengeAuf (1, anzNT);
      firstM : MengenArray (1 .. anzNT) := 
                  (others => LeerMengeAuf (anzNT+1, anzV+1));
      follow : MengenArray (1 .. anzNT) := 
                  (others => LeerMengeAuf (anzNT+1, anzV+1));
   end record;
   
   type Grammatik is access Grammatik_Rec;
   
   ----------------------------------------------------------------

   function txt (Zeichen : Positive; G : Grammatik) return String;
   
   procedure drucke (G : Grammatik);
   
   ---------------------------------------------------------------- 
   
   -- Die GrammatikInfos umfassen:
   --    + die Menge der epsilon-Symbole
   --    + je Nichtterminal A die Mengen first(A) und follow(A)
   --    + je Alternative die zugehrigen LL(1)-Vorschaumengen
   
   procedure bestimmeInfos (G : in out Grammatik;
                                     ablauf : Boolean := False);
   procedure druckeInfos (G : Grammatik);
   

end Grammatiken;